# -*- coding: utf-8 -*-
#
# Copyright (C) 2006 Alec Thomas <alec@swapoff.org>
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import re
from trac.core import *
from tractags.api import TagSystem, ITagProvider
from trac.ticket.model import Ticket
from trac.util.text import to_unicode
from trac.util.compat import set, sorted
from trac.config import *
from trac.resource import Resource
from trac.web.chrome import add_warning


class TicketTagProvider(Component):
    """A tag provider using ticket fields as sources of tags.

    Currently does NOT support custom fields.
    """
    implements(ITagProvider)

    fields = ListOption('tags', 'ticket_fields', 'keywords',
        doc='List of ticket fields to expose as tags.')

    ignore_closed_tickets = BoolOption('tags', 'ignore_closed_tickets', True,
        'Do not collect tags from closed tickets.')

#    custom_fields = ListOption('tags', 'custom_ticket_fields',
#        doc='List of custom ticket fields to expose as tags.')

    # ITagProvider methods
    def get_taggable_realm(self):
        return 'ticket'

    def get_tagged_resources(self, req, tags):
        if 'TICKET_VIEW' not in req.perm:
            return
        # 2008-12-02 added by george, start to deal with custom fields
        standard_fields = ['keywords','component','type','version','milestone','status','resolution','priority','severity']
        s_fields = []
        c_fields = []
        for field in self.fields:
            if field in standard_fields:
                s_fields.append(field)
            else:
                c_fields.append(field)
        # now custom field seperated with standard fields
        split_into_tags = TagSystem(self.env).split_into_tags
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        args = []
        ignore = ''
        if self.ignore_closed_tickets:
            ignore = " WHERE status != 'closed'"
        # prepare clause for custom fields
        left_join_clause = ''
        field_join_clause = ''
        for field in c_fields:
            left_join_clause += " LEFT OUTER JOIN ticket_custom %s ON (t.id = %s.ticket AND %s.name = \'%s\') " % (field, field, field, field)
            field_join_clause += ", %s.value " % (field)
#            field_join_clause += ", %s.value as %s " % (field, field)
#        add_warning(req, left_join_clause)
#        add_warning(req, field_join_clause)
        if s_fields:
            comma_connector = ','
        else:
            comma_connector = ''
        if s_fields and c_fields:
            or_connector = '||'
        else:
            or_connector = ''
#        sql = "SELECT * FROM (SELECT id, %s %s ,%s AS s_fields, %s AS c_fields FROM ticket t %s %s) s" % (
        if self.fields:
            as_fields_clause = ',%s %s %s AS fields' % (
                '||'.join(["COALESCE(%s, '')" % f for f in s_fields]), or_connector,
                '||'.join(["COALESCE(%s.value, '')" % f for f in c_fields]))
        else:
            as_fields_clause = ''
                
        sql = "SELECT * FROM (SELECT id %s %s %s %s FROM ticket t %s %s) s" % (
            comma_connector, ','.join(s_fields), field_join_clause, as_fields_clause,
#            '||'.join(["COALESCE(%s, '')" % f for f in self.fields]),
            left_join_clause,
            ignore)
        constraints = []
        if tags:
            if s_fields and c_fields:
                or_connector = ' OR '
            else:
                or_connector = ''
            if self.fields:
                constraints.append(
                    "(" + ' OR '.join(["fields LIKE %s" for tt in tags]) + ")")
#                    "(" + ' OR '.join(["s_fields LIKE %s" for tt in tags]) + ")")
#                    "(" + ' OR '.join(["s_fields LIKE %s" for tt in tags]) + or_connector + ' OR '.join(["c_fields LIKE %s" for t in tags]) + ")")
                args += ['%' + t + '%' for t in tags]
        elif self.fields:
            constraints.append("fields != ''")
#        add_warning(req, '%s' % sql)
        if constraints:
            sql += " WHERE " + " AND ".join(constraints)
#        raise TracError(constraints)
        sql += " ORDER BY id"
#        raise TracError(args)
#        add_warning(req, '%s' % sql)
        self.env.log.debug(sql)
        cursor.execute(sql, args)
        for row in cursor:
            id, ttags = row[0], ' '.join([f for f in row[1:-1] if f])
            perm = req.perm('ticket', id)
            if 'TICKET_VIEW' not in perm or 'TAGS_VIEW' not in perm:
                continue
            ticket_tags = split_into_tags(ttags)
#            add_warning(req, '[george in debugging] %s %s' % (id, ticket_tags))
# 2008-12-02 george modified to ignore case in query
            temp_ticket_tags = set([to_unicode(x.lower()) for x in ticket_tags])
            tags = set([to_unicode(x.lower()) for x in tags])
#            if (not tags or ticket_tags.intersection(tags)):
            if (not tags or temp_ticket_tags.intersection(tags)):
                yield Resource('ticket', id), ticket_tags
#                add_warning(req, '[george in debugging] %s' % ticket_tags)


    def get_resource_tags(self, req, resource):
        if 'TICKET_VIEW' not in req.perm(resource):
            return
        ticket = Ticket(self.env, resource.id)
        return self._ticket_tags(ticket)

    def set_resource_tags(self, req, resource, tags):
        req.perm.require('TICKET_MODIFY', resource)
        split_into_tags = TagSystem(self.env).split_into_tags
        ticket = Ticket(self.env, resource.id)
        all = self._ticket_tags(ticket)
        keywords = split_into_tags(ticket['keywords'])
        tags.difference_update(all.difference(keywords))
        ticket['keywords'] = u' '.join(sorted(map(to_unicode, tags)))
        ticket.save_changes(req.username, u'')

    def remove_resource_tags(self, req, resource):
        req.perm.require('TICKET_MODIFY', resource)
        ticket = Ticket(self.env, resource.id)
        ticket['keywords'] = u''
        ticket.save_changes(req.username, u'')

    # Private methods
    def _ticket_tags(self, ticket):
        split_into_tags = TagSystem(self.env).split_into_tags
        return split_into_tags(
            ' '.join(filter(None, [ticket[f] for f in self.fields])))
